;
; atan(y/x)߂
;
; 
; DE:W
;
; o
; A:px
; F:j
;
atan::
	push BC
	push DE
	push HL
	call _atan
	pop HL
	pop DE
	pop BC
	ret

_atan:
	; if x = 0 and y = 0 then return
	ld A, D
	or E
	ret Z

	; if x < 0 goto atan_xneg else atan_xpos
	ld A, E
	rla
	jr C, atan_xneg

atan_xpos:
	; if y < 0 goto atan_xpos_yneg else atan_xpos_ypos
	ld A, D
	rla
	jr C, atan_xpos_yneg
	
atan_xpos_ypos:
	; return atan0(x, y)
	jr atan0

atan_xpos_yneg:
	; return -atan0(x, -y)
	rra
	neg
	ld D, A
	call atan0
	neg
	ret

atan_xneg:
	; if y < 0 goto atan_xneg_yneg else atan_xneg_ypos
	rra
	neg
	ld E, A
	ld A, D
	rla
	jr C, atan_xneg_yneg
	
atan_xneg_ypos:
	; return 0x80 - atan0(-x, y)
	call atan0
	neg
	add #0x80
	ret

atan_xneg_yneg:
	; return -(0x80 - atan0(-x, -y))
	rra
	neg
	ld D, A
	call atan0
	neg
	add #0x80
	neg
	ret

atan0:
	; if x < y then swap(x, y)
	push DE
	ld A, E
	cp D
	jr NC, atan0_skip
	ld A, D
	ld D, E
	ld E, A
atan0_skip:

	; dx = x, dy = y, if x >= 64 goto atan0_loop2
	ld B, D
	ld C, E
	ld A, E
	rla
	rla
	jr C, atan0_loop2
	
	; while(x < 64) { x <<= 1, y <<= 1, dx = x, dy = y }
atan0_loop1:
	sla D
	sla E
	ld B, D
	ld C, E
	ld A, E
	rla
	rla
	jr NC, atan0_loop1

atan0_loop2:
	; dx <<= 1, dy <<= 1, if dx = 0 goto atan0_last
	srl C
	jr Z, atan0_last
	srl B
	
	; if x = 63 goto atan0_last
	ld A, E
	cp #63
	jr Z, atan0_last
	
	; if x < 63 goto atan0_back else atan0_fore
	jr NC, atan0_back

atan0_fore:
	; x += dx, y += dy, goto atan0_loop2
	ld A, E
	add C
	ld E, A
	ld A, D
	add B
	ld D, A
	jr atan0_loop2

atan0_back:
	; x -= dx, y -= dy, goto atan0_loop2
	ld A, E
	sub C
	ld E, A
	ld A, D
	sub B
	ld D, A
	jr atan0_loop2

atan0_last:
	; r = atan_table[x]
	ld E, D
	ld D, #0
	ld HL, #atan_table
	add HL, DE
	ld B, (HL)

	; if x > y then return r
	pop DE
	ld A, E
	cp D
	ld A, B
	ret NC

	; return 0x40 - r
	ld A, #0x40
	sub B
	ret

atan_table:
	.db 0x0	; atan(0/63)
	.db 0x0	; atan(1/63)
	.db 0x1	; atan(2/63)
	.db 0x1	; atan(3/63)
	.db 0x2	; atan(4/63)
	.db 0x3	; atan(5/63)
	.db 0x3	; atan(6/63)
	.db 0x4	; atan(7/63)
	.db 0x5	; atan(8/63)
	.db 0x5	; atan(9/63)
	.db 0x6	; atan(10/63)
	.db 0x7	; atan(11/63)
	.db 0x7	; atan(12/63)
	.db 0x8	; atan(13/63)
	.db 0x8	; atan(14/63)
	.db 0x9	; atan(15/63)
	.db 0xa	; atan(16/63)
	.db 0xa	; atan(17/63)
	.db 0xb	; atan(18/63)
	.db 0xb	; atan(19/63)
	.db 0xc	; atan(20/63)
	.db 0xd	; atan(21/63)
	.db 0xd	; atan(22/63)
	.db 0xe	; atan(23/63)
	.db 0xe	; atan(24/63)
	.db 0xf	; atan(25/63)
	.db 0xf	; atan(26/63)
	.db 0x10	; atan(27/63)
	.db 0x11	; atan(28/63)
	.db 0x11	; atan(29/63)
	.db 0x12	; atan(30/63)
	.db 0x12	; atan(31/63)
	.db 0x13	; atan(32/63)
	.db 0x13	; atan(33/63)
	.db 0x14	; atan(34/63)
	.db 0x14	; atan(35/63)
	.db 0x15	; atan(36/63)
	.db 0x15	; atan(37/63)
	.db 0x16	; atan(38/63)
	.db 0x16	; atan(39/63)
	.db 0x17	; atan(40/63)
	.db 0x17	; atan(41/63)
	.db 0x17	; atan(42/63)
	.db 0x18	; atan(43/63)
	.db 0x18	; atan(44/63)
	.db 0x19	; atan(45/63)
	.db 0x19	; atan(46/63)
	.db 0x1a	; atan(47/63)
	.db 0x1a	; atan(48/63)
	.db 0x1a	; atan(49/63)
	.db 0x1b	; atan(50/63)
	.db 0x1b	; atan(51/63)
	.db 0x1c	; atan(52/63)
	.db 0x1c	; atan(53/63)
	.db 0x1c	; atan(54/63)
	.db 0x1d	; atan(55/63)
	.db 0x1d	; atan(56/63)
	.db 0x1d	; atan(57/63)
	.db 0x1e	; atan(58/63)
	.db 0x1e	; atan(59/63)
	.db 0x1f	; atan(60/63)
	.db 0x1f	; atan(61/63)
	.db 0x1f	; atan(62/63)
	.db 0x20	; atan(63/63)
	.db 0x20	; atan(64/63)

;
; Copyright 2011 maruhiro
; All rights reserved. 
;
; Redistribution and use in source and binary forms, 
; with or without modification, are permitted provided that 
; the following conditions are met: 
;
;  1. Redistributions of source code must retain the above copyright notice, 
;     this list of conditions and the following disclaimer. 
;
;  2. Redistributions in binary form must reproduce the above copyright notice, 
;     this list of conditions and the following disclaimer in the documentation 
;     and/or other materials provided with the distribution. 
;
; THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
; INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
; FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
; THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
; ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;

; eof
